import datetime

from sqlalchemy.orm import Mapped

from ziplime.core.db.annotated_types import DateIndexed, IntegerIndexed, StringPK
from ziplime.core.db.base_model import BaseModel


class DividendPayout(BaseModel):
    __tablename__ = "dividend_payouts"
    index: Mapped[StringPK]
    sid: Mapped[IntegerIndexed]
    ex_date: Mapped[DateIndexed]
    declared_date: Mapped[datetime.date]
    record_date: Mapped[datetime.date]
    pay_date: Mapped[datetime.date]
    amount: Mapped[int]



import datetime
from typing import Any

import pandas as pd
from exchange_calendars import ExchangeCalendar

from ziplime.exchanges.exchange import Exchange
from ziplime.finance.domain.ledger import Ledger
from ziplime.sources.benchmark_source import BenchmarkSource


class NumTradingDays:
    """Report the number of trading days."""

    def start_of_simulation(self, ledger: Ledger, emission_rate: datetime.timedelta, trading_calendar: ExchangeCalendar,
                            sessions: pd.DatetimeIndex, benchmark_source: BenchmarkSource):
        self._num_trading_days = 0

    def start_of_session(self, ledger, session, exchanges: dict[str, Exchange]):
        self._num_trading_days += 1

    def end_of_bar(self, packet: dict[str, Any], ledger: Ledger, session: datetime.datetime, session_ix: int,
                   exchanges: dict[str, Exchange]):
        packet["cumulative_risk_metrics"]["trading_days"] = self._num_trading_days

    end_of_session = end_of_bar



import datetime
from collections import OrderedDict
from functools import partial
from math import isnan, copysign

import numpy as np
import structlog

from ziplime.assets.entities.asset import Asset
from ziplime.assets.models.dividend import Dividend
from ziplime.assets.entities.futures_contract import FuturesContract
from ziplime.exchanges.exchange import Exchange
from ziplime.finance.domain.position import Position
from ziplime.finance.domain.transaction import Transaction
from ziplime.finance.finance_ext import (
    PositionStats,
    calculate_position_tracker_stats
)


class PositionTracker:
    """The current state of the positions held.

    Parameters
    ----------
    data_frequency : {'daily', 'minute'}
        The data frequency of the simulation.
    """

    def __init__(self, exchanges: dict[str, Exchange], data_frequency: datetime.timedelta):
        self.positions = OrderedDict()

        self._unpaid_dividends = {}
        self._unpaid_stock_dividends = {}
        self._positions_store = {}

        self.data_frequency = data_frequency
        # self.data_bundle = data_bundle
        # cache the stats until something alters our positions
        self._dirty_stats = True
        self.exchanges = exchanges
        self._stats = PositionStats.new()
        self._logger = structlog.get_logger(__name__)

    def update_position(
            self,
            asset: Asset,
            amount: float | None = None,
            last_sale_price: float | None = None,
            last_sale_date=None,
            cost_basis=None,
    ):
        self._dirty_stats = True

        if asset not in self.positions:
            position = Position(asset,
                                amount=0,
                                cost_basis=float(0.0),
                                last_sale_price=float(0.0),
                                last_sale_date=None)
            self.positions[asset] = position
        else:
            position = self.positions[asset]

        if amount is not None:
            position.amount = amount
        if last_sale_price is not None:
            position.last_sale_price = last_sale_price
        if last_sale_date is not None:
            position.last_sale_date = last_sale_date
        if cost_basis is not None:
            position.cost_basis = cost_basis

    def execute_transaction(self, txn):
        self._dirty_stats = True

        asset = txn.asset

        if asset not in self.positions:
            position = Position(asset=asset,
                                amount=0,
                                cost_basis=0.0,
                                last_sale_price=0.0,
                                last_sale_date=None
                                )
            self.positions[asset] = position
        else:
            position = self.positions[asset]

        self._update_position(position=position, txn=txn)

        if position.amount == 0:
            del self.positions[asset]

            try:
                # if this position exists in our user-facing dictionary,
                # remove it as well.
                del self._positions_store[asset]
            except KeyError:
                pass

    def _update_position(self, position: Position, txn: Transaction):
        if position.asset != txn.asset:
            raise Exception("updating position with txn for a " "different asset")

        total_shares = position.amount + txn.amount

        if total_shares == 0:
            position.cost_basis = 0.0
        else:
            prev_direction = copysign(1, position.amount)
            txn_direction = copysign(1, txn.amount)

            if prev_direction != txn_direction:
                # we're covering a short or closing a position
                if abs(txn.amount) > abs(position.amount):
                    # we've closed the position and gone short
                    # or covered the short position and gone long
                    position.cost_basis = txn.price
            else:
                prev_cost = position.cost_basis * position.amount
                txn_cost = txn.amount * txn.price
                total_cost = prev_cost + txn_cost
                position.cost_basis = total_cost / total_shares

            # Update the last sale price if txn is
            # best data we have so far
            if position.last_sale_date is None or txn.dt > position.last_sale_date:
                position.last_sale_price = txn.price
                position.last_sale_date = txn.dt

        position.amount = total_shares

    def handle_commission(self, asset: Asset, cost: float) -> None:
        # Adjust the cost basis of the stock if we own it
        if asset in self.positions:
            self._dirty_stats = True
            self.adjust_commission_cost_basis(position=self.positions[asset], asset=asset, cost=cost)

    def adjust_commission_cost_basis(self, position: Position, asset: Asset, cost: float):
        """
        A note about cost-basis in ziplime: all positions are considered
        to share a cost basis, even if they were executed in different
        transactions with different commission costs, different prices, etc.

        Due to limitations about how ziplime handles positions, ziplime will
        currently spread an externally-delivered commission charge across
        all shares in a position.
        """

        if asset != position.asset:
            raise Exception("Updating a commission for a different asset?")
        if cost == 0.0:
            return

        # If we no longer hold this position, there is no cost basis to
        # adjust.
        if position.amount == 0:
            return

        # We treat cost basis as the share price where we have broken even.
        # For longs, commissions cause a relatively straight forward increase
        # in the cost basis.
        #
        # For shorts, you actually want to decrease the cost basis because you
        # break even and earn a profit when the share price decreases.
        #
        # Shorts are represented as having a negative `amount`.
        #
        # The multiplication and division by `amount` cancel out leaving the
        # cost_basis positive, while subtracting the commission.

        prev_cost = position.cost_basis * position.amount
        if isinstance(asset, FuturesContract):
            cost_to_use = cost / asset.price_multiplier
        else:
            cost_to_use = cost
        new_cost = prev_cost + cost_to_use
        position.cost_basis = new_cost / position.amount

    def handle_splits(self, splits):
        """Processes a list of splits by modifying any positions as needed.

        Parameters
        ----------
        splits: list
            A list of splits.  Each split is a tuple of (asset, ratio).

        Returns
        -------
        int: The leftover cash from fractional shares after modifying each
            position.
        """
        total_leftover_cash = 0

        for asset, ratio in splits:
            if asset in self.positions:
                self._dirty_stats = True

                # Make the position object handle the split. It returns the
                # leftover cash from a fractional share, if there is any.
                position = self.positions[asset]
                leftover_cash = self.handle_split(position=position, asset=asset, ratio=ratio)
                total_leftover_cash += leftover_cash

        return total_leftover_cash

    def earn_dividend(self, position: Position, dividend: Dividend) -> dict[str, float]:
        """
        Register the number of shares we held at this dividend's ex date so
        that we can pay out the correct amount on the dividend's pay date.
        """
        return {"amount": position.amount * dividend.amount}

    def earn_stock_dividend(self, position: Position, stock_dividend):
        """
        Register the number of shares we held at this dividend's ex date so
        that we can pay out the correct amount on the dividend's pay date.
        """
        return {
            "payment_asset": stock_dividend.payment_asset,
            "share_count": np.floor(position.amount * float(stock_dividend.ratio)),
        }

    def handle_split(self, position: Position, asset: Asset, ratio: float):
        """
        Update the position by the split ratio, and return the resulting
        fractional share that will be converted into cash.

        Returns the unused cash.
        """
        if position.asset != asset:
            raise Exception("updating split with the wrong asset!")

        # adjust the # of shares by the ratio
        # (if we had 100 shares, and the ratio is 3,
        #  we now have 33 shares)
        # (old_share_count / ratio = new_share_count)
        # (old_price * ratio = new_price)

        # e.g., 33.333
        raw_share_count = position.amount / float(ratio)

        # e.g., 33
        full_share_count = np.floor(raw_share_count)

        # e.g., 0.333
        fractional_share_count = raw_share_count - full_share_count

        # adjust the cost basis to the nearest cent, e.g., 60.0
        new_cost_basis = round(position.cost_basis * ratio, 2)

        position.cost_basis = new_cost_basis
        position.amount = full_share_count

        return_cash = round(float(fractional_share_count * new_cost_basis), 2)

        self._logger.info("after split: " + str(position))
        self._logger.info("returning cash: " + str(return_cash))

        # return the leftover cash, which will be converted into cash
        # (rounded to the nearest cent)
        return return_cash

    def earn_dividends(self, cash_dividends, stock_dividends):
        """Given a list of dividends whose ex_dates are all the next trading
        day, calculate and store the cash and/or stock payments to be paid on
        each dividend's pay date.

        Parameters
        ----------
        cash_dividends : iterable of (asset, amount, pay_date) namedtuples

        stock_dividends: iterable of (asset, payment_asset, ratio, pay_date)
            namedtuples.
        """
        for cash_dividend in cash_dividends:
            self._dirty_stats = True  # only mark dirty if we pay a dividend

            # Store the earned dividends so that they can be paid on the
            # dividends' pay_dates.
            div_owed = self.earn_dividend(position=self.positions[cash_dividend.asset], dividend=cash_dividend)
            try:
                self._unpaid_dividends[cash_dividend.pay_date].append(div_owed)
            except KeyError:
                self._unpaid_dividends[cash_dividend.pay_date] = [div_owed]

        for stock_dividend in stock_dividends:
            self._dirty_stats = True  # only mark dirty if we pay a dividend

            div_owed = self.earn_stock_dividend(position=self.positions[stock_dividend.asset],
                                                stock_dividend=stock_dividend)
            try:
                self._unpaid_stock_dividends[stock_dividend.pay_date].append(
                    div_owed,
                )
            except KeyError:
                self._unpaid_stock_dividends[stock_dividend.pay_date] = [
                    div_owed,
                ]

    def pay_dividends(self, next_trading_day: datetime.datetime):
        """
        Returns a cash payment based on the dividends that should be paid out
        according to the accumulated bookkeeping of earned, unpaid, and stock
        dividends.
        """
        net_cash_payment = 0.0

        try:
            payments = self._unpaid_dividends[next_trading_day]
            # Mark these dividends as paid by dropping them from our unpaid
            del self._unpaid_dividends[next_trading_day]
        except KeyError:
            payments = []

        # representing the fact that we're required to reimburse the owner of
        # the stock for any dividends paid while borrowing.
        for payment in payments:
            net_cash_payment += payment["amount"]

        # Add stock for any stock dividends paid.  Again, the values here may
        # be negative in the case of short positions.
        try:
            stock_payments = self._unpaid_stock_dividends[next_trading_day]
        except KeyError:
            stock_payments = []

        for stock_payment in stock_payments:
            payment_asset = stock_payment["payment_asset"]
            share_count = stock_payment["share_count"]
            # note we create a Position for stock dividend if we don't
            # already own the asset
            if payment_asset in self.positions:
                position = self.positions[payment_asset]
            else:
                position = self.positions[payment_asset] = Position(
                    asset=payment_asset,
                    amount=0,
                    cost_basis=0.0,
                    last_sale_price=0.0,
                    last_sale_date=None

                )

            position.amount += share_count

        return net_cash_payment

    def maybe_create_close_position_transaction(self, asset: Asset, dt: datetime.datetime):
        if not self.positions.get(asset):
            return None

        amount = self.positions.get(asset).amount
        # TODO: check this
        price = self.data_bundle.get_spot_value(assets=asset, field="price", dt=dt,
                                                data_frequency=self._data_frequency)

        # Get the last traded price if price is no longer available
        if isnan(price):
            price = self.positions.get(asset).last_sale_price

        return Transaction(
            asset=asset,
            amount=-amount,
            dt=dt,
            price=price,
            order_id=None,
            exchange_name=None
        )

    def get_positions(self):
        for asset, pos in self.positions.items():
            # Adds the new position if we didn't have one before, or overwrite
            # one we have currently
            self._positions_store[asset] = pos

        return self._positions_store

    def get_position_list(self):
        return [
            pos.to_dict() for asset, pos in self.positions.items() if pos.amount != 0
        ]

    def sync_last_sale_prices(self, dt: datetime.datetime,
                              exchange_name:str,
                              handle_non_market_minutes: bool = False):
        self._dirty_stats = True
        exchange = self.exchanges[exchange_name]
        if handle_non_market_minutes:
            previous_minute = exchange.trading_calendar.previous_minute(minute=dt)
            get_price = partial(
                exchange.get_adjusted_value,
                field="close",
                dt=previous_minute,
                perspective_dt=dt,
                frequency=self.data_frequency
            )

        else:
            get_price = partial(
                exchange.get_scalar_asset_spot_value_sync,
                field="close",
                dt=dt,
                frequency=self.data_frequency
            )
        for outer_position in self.positions.values():
            inner_position = outer_position

            last_sale_price = get_price(inner_position.asset)["close"][0]

            # inline ~isnan because this gets called once per position per minute
            if last_sale_price is None:

                self._logger.warning(
                    f"Error updating last sale price for {inner_position.asset.asset_name} on {dt}. Price is None")
            else:  # last_sale_price == last_sale_price:
                inner_position.last_sale_price = last_sale_price
                inner_position.last_sale_date = dt

    @property
    def stats(self):
        """The current status of the positions.

        Returns
        -------
        stats : PositionStats
            The current stats position stats.

        Notes
        -----
        This is cached, repeated access will not recompute the stats until
        the stats may have changed.
        """
        if self._dirty_stats:
            calculate_position_tracker_stats(self.positions, self._stats)
            self._dirty_stats = False

        return self._stats



import datetime
from typing import Callable

import pandas as pd
import polars as pl
from exchange_calendars import ExchangeCalendar

from ziplime.assets.domain.continuous_future import ContinuousFuture
from contextlib import contextmanager
import numpy as np

from ziplime.assets.entities.asset import Asset
from ziplime.constants.period import Period
from ziplime.data.domain.data_bundle import DataBundle
from ziplime.data.services.data_source import DataSource
from ziplime.exchanges.exchange import Exchange


@contextmanager
def handle_non_market_minutes(bar_data):
    try:
        bar_data._handle_non_market_minutes = True
        yield
    finally:
        bar_data._handle_non_market_minutes = False


class BarData:
    """Provides methods for accessing minutely and daily price/volume data from
    Algorithm API functions.

    Also provides utility methods to determine if an asset is alive, and if it
    has recent trade data.

    An instance of this object is passed as ``data`` to
    :func:`~ziplime.api.handle_data` and
    :func:`~ziplime.api.before_trading_start`.

    Parameters
    ----------
    data_bundle : DataBundle
        Provider for bar pricing data.
    simulation_dt_func : callable
        Function which returns the current simulation time.
        This is usually bound to a method of TradingSimulation.
    data_frequency : {'minute', 'daily'}
        The frequency of the bar data; i.e. whether the data is
        daily or minute bars
    restrictions : ziplime.finance.asset_restrictions.Restrictions
        Object that combines and returns restricted list information from
        multiple sources
    """

    def __init__(self,
                 data_sources: dict[str, DataSource],
                 simulation_dt_func: Callable,
                 trading_calendar: ExchangeCalendar,
                 restrictions):
        # self.data_bundle = data_bundle
        self.simulation_dt_func = simulation_dt_func

        # self._daily_mode = (self.data_bundle == "daily")

        self._adjust_minutes = False

        self._trading_calendar = trading_calendar
        self._is_restricted = restrictions.is_restricted
        self.data_sources = data_sources
        self.default_data_source = data_sources[list(data_sources.keys())[0]]
        # first_exchange = exchanges[list(exchanges.keys())[0]]
        # self.default_exchange = first_exchange

    def _get_current_minute(self):
        """Internal utility method to get the current simulation time.

        Possible answers are:
        - whatever the algorithm's get_datetime() method returns (this is what
            `self.simulation_dt_func()` points to)
        - sometimes we're knowingly not in a market minute, like if we're in
            before_trading_start.  In that case, `self._adjust_minutes` is
            True, and we get the previous market minute.
        - if we're in daily mode, get the session label for this minute.
        """
        dt = self.simulation_dt_func()

        if self._adjust_minutes:
            dt = self._trading_calendar.previous_minute(dt)

        # TODO: check this, is it different for daily?
        #
        # if self._daily_mode:
        #     # if we're in daily mode, take the given dt (which is the last
        #     # minute of the session) and get the session label for it.
        #     dt = self.data_portal.trading_calendar.minute_to_session(dt)

        # return dt
        return dt

    def current(self, assets: list[Asset], fields: list[str],
                data_source: str | None = None) -> pl.DataFrame:
        """Returns the "current" value of the given fields for the given assets
        at the current simulation time.

        Parameters
        ----------
        assets : ziplime.assets.Asset or iterable of ziplime.assets.Asset
            The asset(s) for which data is requested.
        fields : str or iterable[str].
            Requested data field(s). Valid field names are: "price",
            "last_traded", "open", "high", "low", "close", and "volume".

        Returns
        -------
        current_value : Scalar, pandas Series, or pandas DataFrame.
            See notes below.

        Notes
        -----
        The return type of this function depends on the types of its inputs:

        - If a single asset and a single field are requested, the returned
          value is a scalar (either a float or a ``datetime.datetime`` depending on
          the field).

        - If a single asset and a list of fields are requested, the returned
          value is a :class:`pd.Series` whose indices are the requested fields.

        - If a list of assets and a single field are requested, the returned
          value is a :class:`pd.Series` whose indices are the assets.

        - If a list of assets and a list of fields are requested, the returned
          value is a :class:`pd.DataFrame`.  The columns of the returned frame
          will be the requested fields, and the index of the frame will be the
          requested assets.

        The values produced for ``fields`` are as follows:

        - Requesting "price" produces the last known close price for the asset,
          forward-filled from an earlier minute if there is no trade this
          minute. If there is no last known value (either because the asset
          has never traded, or because it has delisted) NaN is returned. If a
          value is found, and we had to cross an adjustment boundary (split,
          dividend, etc) to get it, the value is adjusted to the current
          simulation time before being returned.

        - Requesting "open", "high", "low", or "close" produces the open, high,
          low, or close for the current minute. If no trades occurred this
          minute, ``NaN`` is returned.

        - Requesting "volume" produces the trade volume for the current
          minute. If no trades occurred this minute, 0 is returned.

        - Requesting "last_traded" produces the datetime of the last minute in
          which the asset traded, even if the asset has stopped trading. If
          there is no last known value, ``pd.NaT`` is returned.

        If the current simulation time is not a valid market time for an asset,
        we use the most recent market close instead.
        """
        data = {}
        assets = frozenset(assets)
        fields = frozenset(fields)
        if data_source is None:
            data_source = self.default_data_source.name
        if not self._adjust_minutes:
            return self.data_sources[data_source].current(
                assets=assets,
                fields=fields,
                dt=self._get_current_minute(),
            )
        else:
            for field in fields:
                series = pd.Series(data={
                    asset: self.data_sources[data_source].get_adjusted_value(
                        asset,
                        field,
                        self._get_current_minute(),
                        self.simulation_dt_func(),
                        self.data_bundle.frequency
                    )
                    for asset in assets
                }, index=assets, name=field)
                data[field] = series

        return pd.DataFrame(data=data)

    def current_chain(self, continuous_future: ContinuousFuture):
        return self.data_bundle.get_current_future_chain(
            continuous_future=continuous_future,
            dt=self.simulation_dt_func()
        )

    def can_trade(self, assets: list[Asset]):
        """For the given asset or iterable of assets, returns True if all of the
        following are true:

        1. The asset is alive for the session of the current simulation time
           (if current simulation time is not a market minute, we use the next
           session).
        2. The asset's exchange is open at the current simulation time or at
           the simulation calendar's next market minute.
        3. There is a known last price for the asset.

        Parameters
        ----------
        assets: ziplime.assets.Asset or iterable of ziplime.assets.Asset
            Asset(s) for which tradability should be determined.

        Notes
        -----
        The second condition above warrants some further explanation:

        - If the asset's exchange calendar is identical to the simulation
          calendar, then this condition always returns True.
        - If there are market minutes in the simulation calendar outside of
          this asset's exchange's trading hours (for example, if the simulation
          is running on the CMES calendar but the asset is MSFT, which trades
          on the NYSE), during those minutes, this condition will return False
          (for example, 3:15 am Eastern on a weekday, during which the CMES is
          open but the NYSE is closed).

        Returns
        -------
        can_trade : bool or pd.Series[bool]
            Bool or series of bools indicating whether the requested asset(s)
            can be traded in the current minute.
        """
        dt = self.simulation_dt_func()
        assets = set(assets)
        if self._adjust_minutes:
            adjusted_dt = self._get_current_minute()
        else:
            adjusted_dt = dt

        # if isinstance(assets, Asset):
        #     return self._can_trade_for_asset(
        #         assets, dt, adjusted_dt, data_portal
        #     )
        # else:
        tradeable = [
            self._can_trade_for_asset(
                asset=asset, dt=dt, adjusted_dt=adjusted_dt
            )
            for asset in assets
        ]
        return pd.Series(data=tradeable, index=assets, dtype=bool)

    def _can_trade_for_asset(self, asset: Asset, dt: datetime.datetime, adjusted_dt: datetime.datetime) -> bool:
        session_label = None
        dt_to_use_for_exchange_check = None

        if self._is_restricted(assets=frozenset({asset}), dt=adjusted_dt):
            return False

        session_label = self._trading_calendar.minute_to_session(minute=dt)

        if not asset.is_alive_for_session(session_label=session_label):
            # asset isn't alive
            return False

        if asset.auto_close_date and session_label > asset.auto_close_date:
            return False

        # TODO: check this
        _daily_mode = False
        if not _daily_mode:
            # Find the next market minute for this calendar, and check if this
            # asset's exchange is open at that minute.
            if self._trading_calendar.is_open_on_minute(minute=dt):
                dt_to_use_for_exchange_check = dt
            else:
                dt_to_use_for_exchange_check = self._trading_calendar.next_open(minute=dt)

            if not asset.is_exchange_open(dt_minute=dt_to_use_for_exchange_check):
                return False
        # is there a last price?
        return not np.isnan(
            self.data_portal.get_spot_value(
                assets=frozenset({asset}), field="price", dt=adjusted_dt, data_frequency=self.data_frequency
            )
        )

    def history(self, assets: list[Asset], bar_count: int,
                frequency: datetime.timedelta | Period = datetime.timedelta(days=1),
                fields: list[str] | None=None,
                data_source: str | None = None
                ) -> pl.DataFrame:
        """Returns a trailing window of length ``bar_count`` with data for
        the given assets, fields, and frequency, adjusted for splits, dividends,
        and mergers as of the current simulation time.

        The semantics for missing data are identical to the ones described in
        the notes for :meth:`current`.

        Parameters
        ----------
        assets: ziplime.assets.Asset or iterable of ziplime.assets.Asset
            The asset(s) for which data is requested.
        fields: string or iterable of string.
            Requested data field(s). Valid field names are: "price",
            "last_traded", "open", "high", "low", "close", and "volume".
        bar_count: int
            Number of data observations requested.
        frequency: str
            String indicating whether to load daily or minutely data
            observations. Pass '1m' for minutely data, '1d' for daily data.

        Returns
        -------
        history : pd.Series or pd.DataFrame or pd.Panel
            See notes below.

        Notes
        -----
        The return type of this function depends on the types of ``assets`` and
        ``fields``:

        - If a single asset and a single field are requested, the returned
          value is a :class:`pd.Series` of length ``bar_count`` whose index is
          :class:`pd.DatetimeIndex`.

        - If a single asset and multiple fields are requested, the returned
          value is a :class:`pd.DataFrame` with shape
          ``(bar_count, len(fields))``. The frame's index will be a
          :class:`pd.DatetimeIndex`, and its columns will be ``fields``.

        - If multiple assets and a single field are requested, the returned
          value is a :class:`pd.DataFrame` with shape
          ``(bar_count, len(assets))``. The frame's index will be a
          :class:`pd.DatetimeIndex`, and its columns will be ``assets``.

        - If multiple assets and multiple fields are requested, the returned
          value is a :class:`pd.DataFrame` with a pd.MultiIndex containing
          pairs of :class:`pd.DatetimeIndex`, and ``assets``, while the columns
          while contain the field(s). It has shape ``(bar_count * len(assets),
          len(fields))``. The names of the pd.MultiIndex are

              - ``date`` if frequency == '1d'`` or ``date_time`` if frequency == '1m``, and
              - ``asset``

        If the current simulation time is not a valid market time, we use the last market close instead.
        """
        assets = frozenset(assets)
        fields = frozenset(fields) if fields else None

        if data_source is None:
            data_source = self.default_data_source.name

        df = self.data_sources[data_source].get_data_by_limit(assets=assets,
                                                         end_date=self._get_current_minute(),
                                                         limit=bar_count,
                                                         frequency=frequency,
                                                         fields=fields,
                                                         include_end_date=False
                                                         )

        # df = self.exchanges[exchange_name].get_data_by_limit(assets=assets,
        #                                                      end_date=self._get_current_minute(),
        #                                                      limit=bar_count,
        #                                                      frequency=frequency,
        #                                                      fields=fields,
        #                                                      include_end_date=False
        #                                                      )
        if self._adjust_minutes:
            adjs = {
                field: self.exchanges[exchange_name].get_adjustments(
                    assets,
                    field,
                    self._get_current_minute(),
                    self.simulation_dt_func()
                )[0] for field in fields
            }

            df = {
                field: df * adjs[field]
                for field, df in df.items()
            }
        return df

    @property
    def current_dt(self):

        return self.simulation_dt_func()

    @property
    def _handle_non_market_minutes(self):
        return self._adjust_minutes

    @_handle_non_market_minutes.setter
    def _handle_non_market_minutes(self, val):
        self._adjust_minutes = val

    @property
    def current_session(self):
        return self._trading_calendar.minute_to_session(
            self.simulation_dt_func(),
            direction="next"
        )



from ziplime.utils.compat import ExitStack, contextmanager, wraps

from .iface import PipelineHooks, PIPELINE_HOOKS_CONTEXT_MANAGERS
from .no import NoHooks


def delegating_hooks_method(method_name):
    """Factory function for making DelegatingHooks methods."""
    if method_name in PIPELINE_HOOKS_CONTEXT_MANAGERS:
        # Generate a contextmanager that enters the context of all child hooks.
        @wraps(getattr(PipelineHooks, method_name))
        @contextmanager
        def ctx(self, *args, **kwargs):
            with ExitStack() as stack:
                for hook in self._hooks:
                    sub_ctx = getattr(hook, method_name)(*args, **kwargs)
                    stack.enter_context(sub_ctx)
                yield stack

        return ctx
    else:
        # Generate a method that calls methods of all child hooks.
        @wraps(getattr(PipelineHooks, method_name))
        def method(self, *args, **kwargs):
            for hook in self._hooks:
                sub_method = getattr(hook, method_name)
                sub_method(*args, **kwargs)

        return method


class DelegatingHooks(PipelineHooks):
    """A PipelineHooks that delegates to one or more other hooks.

    Parameters
    ----------
    hooks : list[implements(PipelineHooks)]
        Sequence of hooks to delegate to.
    """

    def __new__(cls, hooks):
        if len(hooks) == 0:
            # OPTIMIZATION: Short-circuit to a NoHooks if we don't have any
            # sub-hooks.
            return NoHooks()
        elif len(hooks) == 1:
            # OPTIMIZATION: Unwrap delegation layer if we only have one
            # sub-hook.
            return hooks[0]
        else:
            self = super(DelegatingHooks, cls).__new__(cls)
            self._hooks = hooks
            return self

    # Implement all interface methods by delegating to corresponding methods on
    # input hooks.
    # locals().update(
    #     {
    #         name: delegating_hooks_method(name)
    #         # TODO: Expose this publicly on interface.
    #         for name in PipelineHooks._signatures
    #     }
    # )


del delegating_hooks_method



import math

from ziplime.assets.entities.asset import Asset
from ziplime.finance.domain.order_status import OrderStatus
from ziplime.finance.execution import ExecutionStyle
from ziplime.protocol import DataSourceType

SELL = 1 << 0
BUY = 1 << 1
STOP = 1 << 2
LIMIT = 1 << 3


# ORDER_FIELDS_TO_IGNORE = {"type", "direction", "_status", "asset"}


class Order:
    def __init__(
            self,
            id,
            dt,
            asset: Asset,
            amount: int,
            filled: int,
            commission: float,
            execution_style: ExecutionStyle,
            status: OrderStatus,
            exchange_name: str,
            exchange_order_id: str = None
    ):
        """
        @dt - datetime.datetime that the order was placed
        @asset - asset for the order.
        @amount - the number of shares to buy/sell
                  a positive sign indicates a buy
                  a negative sign indicates a sell
        @filled - how many shares of the order have been filled so far
        """

        # get a string representation of the uuid.
        self.id = id
        self.dt = dt
        self.reason = None
        self.created = dt
        self.asset = asset
        self.amount = amount
        self.filled = filled
        self.commission = commission
        self.exchange_name = exchange_name
        self._status = status

        is_buy = amount > 0
        self.stop = execution_style.get_stop_price(is_buy=is_buy)
        self.limit = execution_style.get_limit_price(is_buy=is_buy)
        self.stop_reached = False
        self.limit_reached = False
        self.direction = math.copysign(1, self.amount)
        self.type = DataSourceType.ORDER
        self.execution_style = execution_style
        self.exchange_order_id = exchange_order_id

    def to_dict(self):
        dct = {
            "amount": self.amount,
            "commission": self.commission,
            "created": self.created,
            "dt": self.dt,
            "exchange_order_id": self.exchange_order_id,
            "filled": self.filled,
            "id": self.id,
            "limit": self.limit,
            "limit_reached": self.limit_reached,
            "reason": self.reason,
            "stop": self.stop,
            "stop_reached": self.stop_reached,
            "status": self.status
        }
        return dct

    def check_triggers(self, price, dt):
        """
        Update internal state based on price triggers and the
        trade event's price.
        """
        (
            stop_reached,
            limit_reached,
            sl_stop_reached,
        ) = self.check_order_triggers(price)
        if (stop_reached, limit_reached) != (
                self.stop_reached,
                self.limit_reached,
        ):
            self.dt = dt
        self.stop_reached = stop_reached
        self.limit_reached = limit_reached
        if sl_stop_reached:
            # Change the STOP LIMIT order into a LIMIT order
            self.stop = None

    def get_order_type(self) -> int:
        order_type = 0

        if self.amount > 0:
            order_type |= BUY
        else:
            order_type |= SELL

        if self.stop is not None:
            order_type |= STOP

        if self.limit is not None:
            order_type |= LIMIT
        return order_type

    # TODO: simplify
    # flake8: noqa: C901
    def check_order_triggers(self, current_price):
        """
        Given an order and a trade event, return a tuple of
        (stop_reached, limit_reached).
        For market orders, will return (False, False).
        For stop orders, limit_reached will always be False.
        For limit orders, stop_reached will always be False.
        For stop limit orders a Boolean is returned to flag
        that the stop has been reached.

        Orders that have been triggered already (price targets reached),
        the order's current values are returned.
        """
        if self.triggered:
            return self.stop_reached, self.limit_reached, False

        stop_reached = False
        limit_reached = False
        sl_stop_reached = False
        order_type = self.get_order_type()
        if order_type == BUY | STOP | LIMIT:
            if current_price >= self.stop:
                sl_stop_reached = True
                if current_price <= self.limit:
                    limit_reached = True
        elif order_type == SELL | STOP | LIMIT:
            if current_price <= self.stop:
                sl_stop_reached = True
                if current_price >= self.limit:
                    limit_reached = True
        elif order_type == BUY | STOP:
            if current_price >= self.stop:
                stop_reached = True
        elif order_type == SELL | STOP:
            if current_price <= self.stop:
                stop_reached = True
        elif order_type == BUY | LIMIT:
            if current_price <= self.limit:
                limit_reached = True
        elif order_type == SELL | LIMIT:
            # This is a SELL LIMIT order
            if current_price >= self.limit:
                limit_reached = True

        return stop_reached, limit_reached, sl_stop_reached

    def handle_split(self, ratio):
        # update the amount, limit_price, and stop_price
        # by the split's ratio

        # info here: http://finra.complinet.com/en/display/display_plain.html?
        # rbid=2403&element_id=8950&record_id=12208&print=1

        # new_share_amount = old_share_amount / ratio
        # new_price = old_price * ratio

        self.amount = int(self.amount / ratio)

        if self.limit is not None:
            self.limit = round(self.limit * ratio, 2)

        if self.stop is not None:
            self.stop = round(self.stop * ratio, 2)

    @property
    def status(self):
        if not self.open_amount:
            return OrderStatus.FILLED
        elif self._status == OrderStatus.HELD and self.filled:
            return OrderStatus.OPEN
        else:
            return self._status

    @status.setter
    def status(self, status):
        self._status = status

    def cancel(self):
        self.status = OrderStatus.CANCELLED

    def reject(self, reason: str = ""):
        self.status = OrderStatus.REJECTED
        self.reason = reason

    def hold(self, reason: str = ""):
        self.status = OrderStatus.HELD
        self.reason = reason

    @property
    def open(self):
        return self.status in [OrderStatus.OPEN, OrderStatus.HELD]

    @property
    def triggered(self) -> bool:
        """
        For a market order, True.
        For a stop order, True IFF stop_reached.
        For a limit order, True IFF limit_reached.
        """
        if self.stop is not None and not self.stop_reached:
            return False

        if self.limit is not None and not self.limit_reached:
            return False

        return True

    @property
    def open_amount(self):
        return self.amount - self.filled

    def __repr__(self):
        """
        String representation for this object.
        """
        return "Order(%s)" % self.to_dict().__repr__()



from typing import Literal

from ziplime.errors import UnsupportedPipelineOutput

from .domain import Domain, GENERIC, infer_domain
from .graph import ExecutionPlan, TermGraph, SCREEN_NAME
from .filters import Filter
from .term import AssetExists, ComputableTerm, Term


class Pipeline:
    """
    A Pipeline object represents a collection of named expressions to be
    compiled and executed by a PipelineEngine.

    A Pipeline has two important attributes: 'columns', a dictionary of named
    :class:`~ziplime.pipeline.Term` instances, and 'screen', a
    :class:`~ziplime.pipeline.Filter` representing criteria for
    including an asset in the results of a Pipeline.

    To compute a pipeline in the context of a TradingAlgorithm, users must call
    ``attach_pipeline`` in their ``initialize`` function to register that the
    pipeline should be computed each trading day. The most recent outputs of an
    attached pipeline can be retrieved by calling ``pipeline_output`` from
    ``handle_data``, ``before_trading_start``, or a scheduled function.

    Parameters
    ----------
    columns : dict, optional
        Initial columns.
    screen : ziplime.pipeline.Filter, optional
        Initial screen.
    """

    __slots__ = ("_columns", "_screen", "_domain", "__weakref__")

    def __init__(self, columns: dict | None = None, screen: Filter | None = None, domain: Domain = GENERIC):
        if columns is None:
            columns = {}

        validate_column = self.validate_column
        for column_name, term in columns.items():
            validate_column(column_name, term)
            if not isinstance(term, ComputableTerm):
                raise TypeError(
                    "Column {column_name!r} contains an invalid pipeline term "
                    "({term}). Did you mean to append '.latest'?".format(
                        column_name=column_name,
                        term=term,
                    )
                )

        self._columns = columns
        self._screen = screen
        self._domain = domain

    @property
    def columns(self):
        """The output columns of this pipeline.

        Returns
        -------
        columns : dict[str, ziplime.pipeline.ComputableTerm]
            Map from column name to expression computing that column's output.
        """
        return self._columns

    @property
    def screen(self):
        """
        The screen of this pipeline.

        Returns
        -------
        screen : ziplime.pipeline.Filter or None
            Term defining the screen for this pipeline. If ``screen`` is a
            filter, rows that do not pass the filter (i.e., rows for which the
            filter computed ``False``) will be dropped from the output of this
            pipeline before returning results.

        Notes
        -----
        Setting a screen on a Pipeline does not change the values produced for
        any rows: it only affects whether a given row is returned. Computing a
        pipeline with a screen is logically equivalent to computing the
        pipeline without the screen and then, as a post-processing-step,
        filtering out any rows for which the screen computed ``False``.
        """
        return self._screen

    def add(self, term: Term, name: str, overwrite: bool = False):
        """Add a column.

        The results of computing ``term`` will show up as a column in the
        DataFrame produced by running this pipeline.

        Parameters
        ----------
        column : ziplime.pipeline.Term
            A Filter, Factor, or Classifier to add to the pipeline.
        name : str
            Name of the column to add.
        overwrite : bool
            Whether to overwrite the existing entry if we already have a column
            named `name`.
        """
        self.validate_column(name, term)

        columns = self.columns
        if name in columns:
            if overwrite:
                self.remove(name)
            else:
                raise KeyError("Column '{}' already exists.".format(name))

        if not isinstance(term, ComputableTerm):
            raise TypeError(
                "{term} is not a valid pipeline column. Did you mean to "
                "append '.latest'?".format(term=term)
            )

        self._columns[name] = term

    def remove(self, name: str):
        """Remove a column.

        Parameters
        ----------
        name : str
            The name of the column to remove.

        Raises
        ------
        KeyError
            If `name` is not in self.columns.

        Returns
        -------
        removed : ziplime.pipeline.Term
            The removed term.
        """
        return self.columns.pop(name)

    def set_screen(self, screen: Filter, overwrite: bool | int):
        """Set a screen on this Pipeline.

        Parameters
        ----------
        filter : ziplime.pipeline.Filter
            The filter to apply as a screen.
        overwrite : bool
            Whether to overwrite any existing screen.  If overwrite is False
            and self.screen is not None, we raise an error.
        """
        if self._screen is not None and not overwrite:
            raise ValueError(
                "set_screen() called with overwrite=False and screen already "
                "set.\n"
                "If you want to apply multiple filters as a screen use "
                "set_screen(filter1 & filter2 & ...).\n"
                "If you want to replace the previous screen with a new one, "
                "use set_screen(new_filter, overwrite=True)."
            )
        self._screen = screen

    def to_execution_plan(self, domain, default_screen, start_date, end_date):
        """
        Compile into an ExecutionPlan.

        Parameters
        ----------
        domain : ziplime.pipeline.domain.Domain
            Domain on which the pipeline will be executed.
        default_screen : ziplime.pipeline.Term
            Term to use as a screen if self.screen is None.
        all_dates : pd.DatetimeIndex
            A calendar of dates to use to calculate starts and ends for each
            term.
        start_date : pd.Timestamp
            The first date of requested output.
        end_date : pd.Timestamp
            The last date of requested output.

        Returns
        -------
        graph : ziplime.pipeline.graph.ExecutionPlan
            Graph encoding term dependencies, including metadata about extra
            row requirements.
        """
        if self._domain is not GENERIC and self._domain is not domain:
            raise AssertionError(
                "Attempted to compile Pipeline with domain {} to execution "
                "plan with different domain {}.".format(self._domain, domain)
            )

        return ExecutionPlan(
            domain=domain,
            terms=self._prepare_graph_terms(default_screen),
            start_date=start_date,
            end_date=end_date,
        )

    def to_simple_graph(self, default_screen):
        """
        Compile into a simple TermGraph with no extra row metadata.

        Parameters
        ----------
        default_screen : ziplime.pipeline.Term
            Term to use as a screen if self.screen is None.

        Returns
        -------
        graph : ziplime.pipeline.graph.TermGraph
            Graph encoding term dependencies.
        """
        return TermGraph(self._prepare_graph_terms(default_screen))

    def _prepare_graph_terms(self, default_screen):
        """Helper for to_graph and to_execution_plan."""
        columns = self.columns.copy()
        screen = self.screen
        if screen is None:
            screen = default_screen
        columns[SCREEN_NAME] = screen
        return columns

    def show_graph(self, format: Literal["svg", "png", "jpeg"]="svg"):
        """
        Render this Pipeline as a DAG.

        Parameters
        ----------
        format : {'svg', 'png', 'jpeg'}
            Image format to render with.  Default is 'svg'.
        """
        g = self.to_simple_graph(AssetExists())
        if format == "svg":
            return g.svg
        elif format == "png":
            return g.png
        elif format == "jpeg":
            return g.jpeg
        else:
            # We should never get here because of the expect_element decorator
            # above.
            raise AssertionError("Unknown graph format %r." % format)

    @staticmethod
    def validate_column(column_name: str, term: Term):
        if term.ndim == 1:
            raise UnsupportedPipelineOutput(column_name=column_name, term=term)

    @property
    def _output_terms(self):
        """
        A list of terms that are outputs of this pipeline.

        Includes all terms registered as data outputs of the pipeline, plus the
        screen, if present.
        """
        terms = list(self._columns.values())
        screen = self.screen
        if screen is not None:
            terms.append(screen)
        return terms

    def domain(self, default: Domain):
        """
        Get the domain for this pipeline.

        - If an explicit domain was provided at construction time, use it.
        - Otherwise, infer a domain from the registered columns.
        - If no domain can be inferred, return ``default``.

        Parameters
        ----------
        default : ziplime.pipeline.domain.Domain
            Domain to use if no domain can be inferred from this pipeline by
            itself.

        Returns
        -------
        domain : ziplime.pipeline.domain.Domain
            The domain for the pipeline.

        Raises
        ------
        AmbiguousDomain
        ValueError
            If the terms in ``self`` conflict with self._domain.
        """
        # Always compute our inferred domain to ensure that it's compatible
        # with our explicit domain.
        inferred = infer_domain(self._output_terms)

        if inferred is GENERIC and self._domain is GENERIC:
            # Both generic. Fall back to default.
            return default
        elif inferred is GENERIC and self._domain is not GENERIC:
            # Use the non-generic domain.
            return self._domain
        elif inferred is not GENERIC and self._domain is GENERIC:
            # Use the non-generic domain.
            return inferred
        else:
            # Both non-generic. They have to match.
            if inferred is not self._domain:
                raise ValueError(
                    "Conflicting domains in Pipeline. Inferred {}, but {} was "
                    "passed at construction.".format(inferred, self._domain)
                )
            return inferred



from abc import ABCMeta
from itertools import chain


class FinancialModelMeta(ABCMeta):
    """
    This metaclass allows users to create custom slippage and commission models
    that support both equities and futures by subclassing the appropriate
    individualized classes.

    Take for example the following custom model, which subclasses both
    EquitySlippageModel and FutureSlippageModel together:

        class MyCustomSlippage(EquitySlippageModel, FutureSlippageModel):
            def process_order(self, data, order):
                ...

    Ordinarily the first parent class in the MRO ('EquitySlippageModel' in this
    example) would override the 'allowed_asset_types' class attribute to only
    include equities. However, this is probably not the expected behavior for a
    reasonable user, so the metaclass will handle this specific case by
    manually allowing both equities and futures for the class being created.
    """

    def __new__(mcls, name, bases, dict_):
        if "allowed_asset_types" not in dict_:
            allowed_asset_types = tuple(
                chain.from_iterable(
                    marker.allowed_asset_types
                    for marker in bases
                    if isinstance(marker, AllowedAssetMarker)
                )
            )
            if allowed_asset_types:
                dict_["allowed_asset_types"] = allowed_asset_types

        return super(FinancialModelMeta, mcls).__new__(
            mcls,
            name,
            bases,
            dict_,
        )


class AllowedAssetMarker(FinancialModelMeta):
    pass



import datetime

from ziplime.assets.entities.asset import Asset
from ziplime.protocol import DataSourceType


class Transaction:

    def __init__(self, asset: Asset, amount: int, dt: datetime.datetime, price: float, order_id: str,
                 exchange_name: str,
                 commission: float | None = None):
        self.asset = asset
        self.amount = amount
        self.exchange_name=exchange_name
        # if amount < 1:
        #     raise Exception("Transaction magnitude must be at least 1.")

        self.dt = dt
        self.price = price
        self.order_id = order_id
        self.type = DataSourceType.TRANSACTION
        self.commission = commission

    def to_dict(self):
        return {
            "type": self.type,
            "amount": self.amount,
            "dt": self.dt,
            "price": self.price,
            "order_id": self.order_id,
            "commission": self.commission,
            "asset": self.asset,
        }



from functools import partial

from numpy import array, iinfo
from pandas import Timestamp

from ziplime.assets.domain.contract_node import ContractNode


def delivery_predicate(codes, contract):
    # This relies on symbols that are construct following a pattern of
    # root symbol + delivery code + year, e.g. PLF16
    # This check would be more robust if the future contract class had
    # a 'delivery_month' member.
    delivery_code = contract.symbol[-3]
    return delivery_code in codes

march_cycle_delivery_predicate = partial(delivery_predicate,
                                         set(['H', 'M', 'U', 'Z']))

CHAIN_PREDICATES = {
    'EL': march_cycle_delivery_predicate,
    'ME': march_cycle_delivery_predicate,
    'PL': partial(delivery_predicate, set(['F', 'J', 'N', 'V'])),
    'PA': march_cycle_delivery_predicate,

    # The majority of trading in these currency futures is done on a
    # March quarterly cycle (Mar, Jun, Sep, Dec) but contracts are
    # listed for the first 3 consecutive months from the present day. We
    # want the continuous futures to be composed of just the quarterly
    # contracts.
    'JY': march_cycle_delivery_predicate,
    'CD': march_cycle_delivery_predicate,
    'AD': march_cycle_delivery_predicate,
    'BP': march_cycle_delivery_predicate,

    # Gold and silver contracts trade on an unusual specific set of months.
    'GC': partial(delivery_predicate, set(['G', 'J', 'M', 'Q', 'V', 'Z'])),
    'XG': partial(delivery_predicate, set(['G', 'J', 'M', 'Q', 'V', 'Z'])),
    'SV': partial(delivery_predicate, set(['H', 'K', 'N', 'U', 'Z'])),
    'YS': partial(delivery_predicate, set(['H', 'K', 'N', 'U', 'Z'])),
}

ADJUSTMENT_STYLES = {'add', 'mul', None}

class OrderedContracts:
    """A container for aligned values of a future contract chain, in sorted order
    of their occurrence.
    Used to get answers about contracts in relation to their auto close
    dates and start dates.

    Members
    -------
    root_symbol : str
        The root symbol of the future contract chain.
    contracts : deque
        The contracts in the chain in order of occurrence.
    start_dates : long[:]
        The start dates of the contracts in the chain.
        Corresponds by index with contract_sids.
    auto_close_dates : long[:]
        The auto close dates of the contracts in the chain.
        Corresponds by index with contract_sids.
    future_chain_predicates : dict
        A dict mapping root symbol to a predicate function which accepts a contract
    as a parameter and returns whether or not the contract should be included in the
    chain.

    Instances of this class are used by the simulation engine, but not
    exposed to the algorithm.
    """

    # cdef readonly object root_symbol
    # cdef readonly object _head_contract
    # cdef readonly dict sid_to_contract
    # cdef readonly int64_t _start_date
    # cdef readonly int64_t _end_date
    # cdef readonly object chain_predicate

    def __init__(self, root_symbol, contracts, chain_predicate=None):

        self.root_symbol = root_symbol

        self.sid_to_contract = {}

        self._start_date = iinfo('int64').max
        self._end_date = 0

        if chain_predicate is None:
            chain_predicate = lambda x: True

        self._head_contract = None
        prev = None
        while contracts:
            contract = contracts.popleft()

            # It is possible that the first contract in our list has a start
            # date on or after its auto close date. In that case the contract
            # is not tradable, so do not include it in the chain.
            if prev is None and contract.start_date >= contract.auto_close_date:
                continue

            if not chain_predicate(contract):
                continue

            self._start_date = min(contract.start_date.value, self._start_date)
            self._end_date = max(contract.end_date.value, self._end_date)

            curr = ContractNode(contract)
            self.sid_to_contract[contract.sid] = curr
            if self._head_contract is None:
                self._head_contract = curr
                prev = curr
                continue
            curr.prev = prev
            prev.next = curr
            prev = curr

    def contract_before_auto_close(self, dt_value):
        """Get the contract with next upcoming auto close date."""
        curr = self._head_contract
        while curr.next is not None:
            if curr.contract.auto_close_date.value > dt_value:
                break
            curr = curr.next
        return curr.contract.sid

    def contract_at_offset(self, sid, offset, start_cap):
        """Get the sid which is the given sid plus the offset distance.
        An offset of 0 should be reflexive.
        """
        # cdef Py_ssize_t i
        curr = self.sid_to_contract[sid]
        i = 0
        while i < offset:
            if curr.next is None:
                return None
            curr = curr.next
            i += 1
        if curr.contract.start_date.value <= start_cap:
            return curr.contract.sid
        else:
            return None

    def active_chain(self,  starting_sid, dt_value):
        curr = self.sid_to_contract[starting_sid]
        contracts = []

        while curr is not None:
            if curr.contract.start_date.value <= dt_value:
                contracts.append(curr.contract.sid)
            curr = curr.next

        return array(contracts, dtype='int64')

    @property
    def start_date(self):
            return Timestamp(self._start_date)

    @property
    def end_date(self):
            return Timestamp(self._end_date)
